/****************************************************************************
 *   Copyright (C) 2004 by Jurij Ivastsuk-Kienbaum                          *
 *   jurij.ivastsuk@t-online.de                                             *
 *                                                                          *
 *   This program is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 2 of the License,         *
 *   or (at your option) any later version.                                 *
 *   This program is distributed in the hope that it will be useful,        *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty            *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                *
 *   See the GNU General Public License for more details.                   *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to the Free Software Foundation,*
 *   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA            *
 ***************************************************************************/

#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <getopt.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/statfs.h>
#include <fcntl.h>
#include "zlib.h"
#include "safe_read.h"
#include "safe_write.h"


using namespace std;

#define DIR_CHAR '/'


#ifndef MAX_PATH_LEN
#  define MAX_PATH_LEN   1024 /* max pathname length */
#endif


int file_count;  	// amount of input files
int ifd;		// input-file descritor
int ofd;		// output-file descritor
int of_count;		// the amount of outputfiles
int opath = 0;		// out-path exists or not
int ipath = 0;		// in-path exists or not
int out_filename = 0;	// output-filename as parameter exists or not
int chunk_size_known = 0; // tells if the chunk-size as parameter delivered or not
int debug = 0;		// if debug == 1, debug information will be written to debug.log file
int noerror = 0;	// switch to signal if the read operation should proceed after I/O error (0 = no; 1 = yes)
int io_block_size = 1024;	// the size of input and/or output block
string m_c; 		/* bytes multiplicator counter,
			   [b] bytes, [M] megabytes, [G] gigabytes
                           or [s] sektors (512 bytes) */
off_t ifile_size;      	//input file size
off_t chunk_size;  		// size of the splited part in bytes
float init_value = 0;		// size of splited part (parameter as float number)
string progname;   	// program name
char ifname[MAX_PATH_LEN]; 	// input file name
int to_stdout = 0;    		/* output to stdout (-c) */
int from_stdin = 0;		/* input from stdin */

string out_filen;	// output filename

string out_pathn;	// output path name
string in_pathn;	// input path name

struct stat istat;         	//status for input file

struct tm *myTime;
time_t TimeNow = time(NULL);

int level = 6;	// zip level (from 1 to 9) 1 - fast 9 - best
		// default value is 6
unsigned long long end_sum_read = 0; 
unsigned long long end_sum_written = 0; 

unsigned long end_sum_elps_time = 0;  
string total_time_elps;	
string av_speed_s;

string cut_path(string path);

string cut_last_slash(string path);

int do_zsplit(int ifd, int ofd, int level, string path_ifn, string path_ofn, off_t chnk_size);

int do_zsplit(int ifd, int ofd, int level, string path_ifn, string path_ofn);

string get_mult(string pattern);

string get_def_filename();

void log(string log_line);

string get_path(string path);

string make_ofname(string init_name, int indx);

void usage();

void help();

int do_stat(const char *name, struct stat *sbuf);

int do_stat(int fd, struct stat *sbuf);

int treat_file(string iname);

int treat_stdin();

void do_exit(int exit_code);

void close_in_out_files(int ifd, int ofd);

uInt seek_on_err(int ifd, size_t count);

string elaps_time(unsigned long time_sec);

string av_speed(long time_sec, unsigned long long total_bytes_read);

 char log_buf[1024];
 string log_str;
 
 string version = "1.0.1";
 string vers_info = "Copyright (C) 2005 Jurij Ivastsuk-Kienbaum.";
 string copyleft = "Written by Jurij Ivastsuk-Kienbaum.\nCopyright (C) 2005 Jurij Ivastsuk-Kienbaum.\nThis program comes with ABSOLUTELY NO WARRANTY.\nThis is free software, and you are welcome to redistribute it under conditions of GNU GPL.";


struct option longopts[] =
{
  /*{name  has_arg *flag val}*/
  {"help",	0, 0, 'h'}, /* help information */
  {"usage",	0, 0, '?'}, /* help, usage information */
  {"debug",	0, 0, 'd'}, /* debug, some information about read and written bytes will be spooled to debug.log file */
  {"stdout",	0, 0, 'c'}, /* write output to standard output */
  {"outdir",	1, 0, 'o'}, /* output directory */
  {"indir",	1, 0, 'i'}, /* input directory */
  {"outname",	1, 0, 'N'}, /* file name which will be used as output file name instead of the input-filename*/
  {"endsize",	1, 0, 's'}, /* chunk size, size of the gziped splited part */
  {"version",	1, 0, 'v'}, /* program version */
  {"fast",	0, 0, '1'}, /* compress faster */
  {"best",	0, 0, '9'}, /* compress better */
  {"noerror",	0, 0, 'r'}, /* 	Ignore i/o errors and continue after read errors; pad every input block with NULs
  				Zero the buffer before reading, so that if we get a read error,
	 			whatever data we are able to read is followed by zeros.
	 			This minimizes data loss. */
  {0, 0, 0, 0}
};

int main(int argc, char ** argv)
 {


 	int rc = 0;
	int res =0;
        int optc;

 	progname = cut_path(argv[0]);

	/* get options */
 	while (( optc = getopt_long(argc,argv,"h?dco:i:N:s:vr0123456789", longopts, (int *)0)) != -1 )
  	{
  		switch (optc)
   		{
  		 case 'N' :
        		out_filen.assign(optarg);
			out_filename = 1;
                        break;
   		 case 's' :
			m_c = get_mult(optarg);
			sscanf(optarg, "%fd", &init_value);

			if(!m_c.compare("k"))
                  	{
				chunk_size = static_cast<off_t>(init_value * 1024);
                  	}
			else if(!m_c.compare("M"))
                  	{
				chunk_size = static_cast<off_t>(init_value * 1024 * 1024);
                  	}
			else if(!m_c.compare("G"))
                  	{
				chunk_size = static_cast<off_t>(init_value * 1024 * 1024 * 1024);
                  	}
			else if(!m_c.compare("s"))
                  	{
				chunk_size = static_cast<off_t>(init_value * 512);
                  	}
			else if(!m_c.compare(""))
                  	{
				chunk_size = 0;
                  	}
                        else
                        {
				chunk_size = 0;
                        }

                        break;
                 case 'o' :
			out_pathn.assign(optarg);
			out_pathn = cut_last_slash(out_pathn);
	    		opath = 1;
                        break;
		 case 'i' :
			in_pathn.assign(optarg);
			in_pathn = cut_last_slash(in_pathn);
	    		ipath = 1;
                        break;
		 case 'c' :
			to_stdout = 1;
                        break;
		 case 'd' :
			debug = 1;
                        break;
		 case 'r' :
			noerror = 1;
                        break;
   		 case 'h' :
		 	help();
    			rc = -1;
    			do_exit(rc);
   		 case '?' :
    		 	usage();
    			rc = -1;
    			do_exit(rc);
		case 'v' :
    		 	fprintf(stdout, "%s version: %s\n%s\n", progname.c_str(), version.c_str(), vers_info.c_str());
    			rc = -1;
    			do_exit(rc);
		case '0': case '1':  case '2':  case '3':  case '4':
		case '5': case '6':  case '7':  case '8':  case '9':
	    		level = optc - '0';
			break;
                 default:
	    		/* Error */
	    		usage();
			rc = -1;
	    		do_exit(rc);
   		}
  	}

        file_count = argc - optind;
 	if ( file_count == 1 )
  	{
		if (chunk_size != 0)
		{
			chunk_size_known = 1;

			if(debug == 1)
			{
                                sprintf(log_buf, "%s: start of operation.", progname.c_str());
				log_str = log_buf;
				log(log_str);
				sprintf(log_buf, "%s: compression level selected: %d.", progname.c_str(), level);
				log_str = log_buf;
				log(log_str);
				if (noerror == 1)
				{
					sprintf(log_buf, "%s: \"--noerror\" option in use, I/O read errors will be savely ignored.", progname.c_str());
					log_str = log_buf;
					log(log_str);
				}
				sprintf(log_buf, "%s: %s will be splitted and compressed, approximate size of chunks in bytes: %lld.", progname.c_str(), argv[optind], chunk_size);
				log_str = log_buf;
				log(log_str);
			}
			else
			{
                                fprintf(stderr,"%s: start of operation.\n", progname.c_str());
				fprintf(stderr,"%s: compression level selected: %d.\n", progname.c_str(), level);
				if (noerror == 1)
				{
                                	fprintf(stderr,"%s: \"--noerror\" option in use, I/O read errors will be savely ignored.\n", progname.c_str());
				}
				fprintf(stderr,"%s: %s will be splitted and compressed, approximate size of chunks in bytes: %lld.\n", progname.c_str(), argv[optind], chunk_size);
			}

		}
		else
		{
			if(debug == 1)
			{
                                sprintf(log_buf, "%s: start of operation", progname.c_str());
				log_str = log_buf;
				log(log_str);
				if (noerror == 1)
				{
					sprintf(log_buf, "%s: noerror option in use, I/O read errors will be savely ignored", progname.c_str());
					log_str = log_buf;
					log(log_str);
				}
				sprintf(log_buf, "%s: file/device-image cannot be splitted, size of chunk is unknown - one chunk image will be created", progname.c_str());
				log_str = log_buf;
				log(log_str);
			}
			else
			{
                                fprintf(stderr,"%s: start of operation\n", progname.c_str());
				if (noerror == 1)
				{
                                	fprintf(stderr,"%s: noerror option in use, I/O read errors will be savely ignored\n", progname.c_str());
				}
				fprintf(stderr,"%s: file/device-image cannot be splitted, the size of chunk is unknown, in bytes: %d\n", progname.c_str());
			}

		}
                res = treat_file(argv[optind]);

		if(res == 0)
		{

			if(debug == 1)
  			{
                                sprintf(log_buf, "%s: Success!...", progname.c_str());
				log_str = log_buf;
				log(log_str);
				sprintf(log_buf, "%s: end of operation\n", progname.c_str());
				log_str = log_buf;
				log(log_str);
  			}
			else
			{
                        	fprintf(stderr,"%s: Success!...\n", progname.c_str());
				fprintf(stderr,"%s: end of operation\n", progname.c_str());
			}

			rc = 0;
	    		do_exit(rc);
		}
		else
		{
			if(debug == 1)
  			{
                                sprintf(log_buf, "%s: not succeeded..., error in operation!", progname.c_str());
				log_str = log_buf;
				log(log_str);
				sprintf(log_buf, "%s: end of operation\n", progname.c_str());
				log_str = log_buf;
				log(log_str);
  			}
			else
			{
				fprintf(stderr,"%s: not succeeded..., error in operation!\n", progname.c_str());
				fprintf(stderr,"%s: end of operation\n", progname.c_str());
			}
			rc = -1;
	    		do_exit(rc);
		}

	}
	else if(file_count == 0) 
	{
		if(debug == 1) 
  		{
			sprintf(log_buf, "%s: no input file/device found, input error!", progname.c_str());
			log_str = log_buf;
			log(log_str);
			usage();
		}
		else
		{
			fprintf(stderr,"%s: no input file/device found, input error!\n", progname.c_str());
			usage();
		}
		rc = -1;
	    	do_exit(rc);
	}
        else 
        {
                if(debug == 1) 
  		{
			sprintf(log_buf, "%s: more then one input file/device found, input error!", progname.c_str());
			log_str = log_buf;
			log(log_str);
			usage();
		}
		else
		{
			fprintf(stderr,"%s: more then one input file/device found, input error!\n", progname.c_str());
			usage();
		}
		rc = -1;
	    	do_exit(rc);
        }

 } // END main


void usage()
 {
        if(debug == 1) 
  	{
                sprintf(log_buf, "%s: usage: zsplit [-h?dvr0123456789] [-i input_dir] [-c stdout] [-o output_dir] [-N outname] [-s chunk_size] [input_file [- stdin]]\n", progname.c_str());
		log_str = log_buf;
		log(log_str);
  	}
	else 
	{
 		fprintf (stderr, "%s: usage: zsplit [-h?dvr0123456789] [-i input_dir] [-c stdout] [-o output_dir] [-N outname] [-s chunk_size] [input_file or [- stdin]]\n", progname.c_str());
	}
 }

void help()
 {
 	static const char help_str[] =
	"zsplit - will read big devices or files and make compressed splitted image chunks of it.\n"
	"\n"
	"usage: zsplit [OPTIONS] SOURCE\n"
	"\n"
	"zsplit can read from SOURCE, which can be an input device, -stdin or -file, compress and split\n"
	"the output image file into chunks of specified size.\n"
	"These chunks could be saved on any medium for later restoring to its origin with a help of unzsplit.\n"
	"\n"
	"SOURCE:\n"
	"can be a file (example: /path_to_file/file), device (example: /dev/hda), partition (example: /dev/hda1) or stdin ( - ).\n"
        "\n"
	"OPTIONS:\n"
	" -c, --stdout:	           Write output to stdout (standard output).\n\n"
	" -d, --debug:		   The debug information will be written into the debug.log file in the current directory.\n"
	"			   If this option is not used, the debug information will be printed to stderr.\n\n"
	" -i, --indir DIR:	   Input directory (DIR is directory which containes the SOURCE).\n\n"
	" -N, --outname OUT_NAME:   This is the common part of file name used for all the split up files\n"
	"           		   if this option is not specified then the input-file's name is\n"
	"           		   used to create this common part. If input-file name could not be used (example: stdin)\n"
	"			   then the default OUT_NAME will be created based on current date and creation time.\n\n"
	" -h, --help:		   Shows this message.\n\n"
	" -o, --outdir DIR:	   Output directory (DIR is directory which containes the output files).\n\n"
	" -r, --noerror:            If during tne read operation I/O error happens, savely ignore these errors and read farther.\n\n"
	" -s, --endsize SIZE:	   Split the file into chunks of this size SIZE.\n"
	"			   Example of SIZE: 1000s (1000 sectors [1s=512 bytes]),\n"
	"		      	   10k (10 KiB), 100M (100 MiB) or 50G (50 GiB).\n"
	"			   If this option is not used the output file will not be splitted.\n\n"
	" -?, --usage:		   Shows short usage information.\n\n"
	" -v, --version:            Prints out a current version number.\n\n"
	" -#, --fast (or --best):   Regulate the speed of compression using the specified digit #,\n"
	"			   where  0  indicates  no  compression,\n"
	"			   1 indicates the fastest (--fast) compression method (less compression)\n"
	"			   and 9 indicates the slowest (--best) compression method. The default  compression level is 6.\n\n"
	"EXAMPLES:\n"
	"zsplit -s 4.5G -N WinXP_bkp -d /dev/hda\n"
	"	zsplit reads device /dev/hda, compresses using default compression method,\n"
	"	splits up the chunks of size 4.5 gibibyte (GiB), saves these chunks using common\n"
	"	name part WinXP_bkp as: WinXP_bkp_0.spl.zp, WinXP_bkp_1.spl.zp ..., and\n"
	"	writes the debug information to the debug.log file in the current directory.\n\n"
	"zsplit -s 100M -N test_img -o /home/user -d -\n"
	"	zsplit reads from stdin, compresses using default compression method,\n"
	"	splits up the chunks of size 100 mebibyte (MiB), saves these chunks using common\n"
	"	name part test_img as: test_img_0.spl.zp, test_img_1.spl.zp ... into the directory\n"
	"	/home/user, and writes the debug information to the debug.log file in the current directory.\n\n"
	"zsplit -N test_bkp /dev/hda\n"
	"	zsplit reads device /dev/hda, compresses using default compression method,\n"
	"	saves the output file as test_bkp_0.spl.zp in the current directory and\n"
	"	writes the debug information to stderr.\n"
	"\n";
 	fprintf(stdout, "%s: version %s\n%s\n\n", progname.c_str(), version.c_str(), copyleft.c_str());
	fprintf(stdout,help_str);
 }

string cut_path(string path)
 {
 	string file;
 	string res = "";
	unsigned int pos;
	file = path;
	pos = file.find_last_of(DIR_CHAR);
	if(pos < string::npos)
	{
		res = file.substr(pos+1);
	}
	else
	{
        	res = file;
	}

  return res;
 }

string cut_last_slash(string path) 
 {
 	unsigned int i;
	string res = "";
	string tmp_path;

	tmp_path = path;
	i = tmp_path.find_last_of(DIR_CHAR);

	if(i+1 == tmp_path.length())
	{
        	tmp_path.erase(i);
		res = tmp_path;
	}
	else
	{
 		res = path;
	}
  return res;
 }

string get_path(string path)
{
 string pathstr;
 unsigned int pos;

	pathstr = path;
	pos = pathstr.find_last_of(DIR_CHAR);
	if(pos < string::npos)
	{
		pathstr = pathstr.substr(0, pos+1);
	}
	else 
	{
        	pathstr = "";
	}

 return pathstr;
}



string make_ofname(string init_name, int indx)
{
 string fname, tmp_name;
 string suffix;
 string prefix;
 string counter;
 ostringstream i_str;

 unsigned int pos, i;

 suffix.assign(".spl.zp");
 prefix.assign("_");
 i_str << indx;
 counter.assign(i_str.str());
 i_str.str("");

 pos = init_name.find_last_of(".");
 if(pos == string::npos)
 {
 	tmp_name = init_name;
 }
 else
 {
 	tmp_name = init_name.substr(0, pos);
 }

 fname = tmp_name + prefix + counter + suffix;

 return fname;
}

string get_mult(string pattern)
{
  	string res = "";
  	unsigned int pos;

  	pos = pattern.find_first_of("QWERTZUIOPASDFGHJKLYXCVBNMqwertzuiopasdfghjklyxcvbnm");

  	if(pos != string::npos)
  	{
    		res = pattern.substr(pos);
  	}
  	else
  	{
    		res = "";
  	}
  return res;
}

string get_def_filename()
{
  string time;
  unsigned int pos;
	time.assign(asctime(localtime(&TimeNow)));
	pos = time.find(" ");
	while(pos < string::npos)
	{
                time.replace(pos, 1, "_");
		pos = time.find(" ", pos+1);
	}
	pos = time.find(":");
	while(pos < string::npos)
	{
                time.replace(pos, 1, "_");
		pos = time.find(":", pos+1);
	}
		
	time = time.erase(time.length()-1);
 return time;
}
		
string av_speed(unsigned long time_sec, unsigned long long total_bytes_read)
{ 
	string ret_av_speed;
	float min;
	float bytes_min;
	int mbytes_min;
	ostringstream av_sp_mb_min;
	
	
	min = time_sec/60;
	bytes_min = total_bytes_read/min;
	
	mbytes_min = (int)(bytes_min/(1024*1024));
	av_sp_mb_min << mbytes_min << "Mb/min";
	ret_av_speed = av_sp_mb_min.str();
	
	return ret_av_speed;
}

string elaps_time(unsigned long time_sec)
{
	string ret_time;
	ostringstream elt_st;
	ostringstream hour, min, sec;
	
	int hr = 0;
	int mn = 0;
	int sc = 0;
		
	if(time_sec > 3599)
	{
		hr = time_sec/3600;
		
		sc = time_sec - (hr*3600);
		
		if(sc > 59)
		{
			mn = sc/60;
			sc = time_sec - (hr*3600) - (mn*60);
		}
		else
		{
			mn = 0;
		}
	}	
	else if (time_sec > 59)
	{	
		mn = time_sec/60;
		sc = time_sec - (mn*60);
	}
	else
	{
		sc = time_sec;
	}
	
	if( hr < 10)
	{
		hour << "0" << hr;
	}
	else
	{
		hour << hr;
	}
	
	if( mn < 10)
	{
		min << "0" << mn;
	}
	else
	{
		min << mn;
	}
	
	if( sc < 10)
	{
		sec << "0" << sc;
	}
	else
	{
		sec << sc;
	}

	elt_st << hour.str() << "hr" << ":" << min.str() << "min" << ":" << sec.str() << "sec";
	
	ret_time = elt_st.str();
   return ret_time;
}

void log(string log_line)
{
  filebuf buffer;
  iostream stream(&buffer);
  time_t CurrTime = time(NULL);
  ostringstream t_st;
  struct tm mytime;
  ostringstream mon, day, hour, min, sec;

   mytime = *localtime(&CurrTime); 

   if(mytime.tm_mon +1 < 10)
   {
   	mon << "0" << (mytime.tm_mon +1);
   }
   else
   {
   	mon << (mytime.tm_mon +1);
   }

   if(mytime.tm_mday < 10)
   {
   	day << "0" << mytime.tm_mday;
   }
   else
   {
   	day << mytime.tm_mday;
   }

   if(mytime.tm_hour < 10)
   {
   	hour << "0" << mytime.tm_hour;
   }
   else
   {
   	hour << mytime.tm_hour;
   }

   if(mytime.tm_min < 10)
   {
   	min << "0" << mytime.tm_min;
   }
   else
   {
   	min << mytime.tm_min;
   }

   if(mytime.tm_sec < 10)
   {
   	sec << "0" << mytime.tm_sec;
   }
   else
   {
   	sec << mytime.tm_sec;
   }

  buffer.open("debug.log", ios::out | ios::app);

  t_st << day.str() << "-" << mon.str() << "-" << (mytime.tm_year+1900) << " " << hour.str() << ":" << min.str() << ":" << sec.str();
  stream << t_st.str() << " " << log_line.c_str() << '\n';

  buffer.close();
}


void close_in_out_files(int ifd, int ofd)
{
  int close_res;
	if(ifd != 0)
	{
		close_res = close(ifd);
		if(close_res != 0)
		{
			if(debug == 1)
			{
                                sprintf(log_buf, "%s: error in closing the input-file, in: %s", progname.c_str(), "close_in_out_files()");
				log_str = log_buf;
				log(log_str);
			}
			else
			{
	  			fprintf(stderr, "%s: error in closing the input-file, in: %s\n", progname.c_str(), "close_in_out_files()");
			}
		}
	  ifd = 0;
	}
	if(ofd != 0)
	{
		close_res = close(ofd);
		if(close_res != 0)
		{
	  		if(debug == 1)
			{
                                sprintf(log_buf, "%s: error in closing the output-file, in: %s", progname.c_str(), "close_in_out_files()");
				log_str = log_buf;
				log(log_str);
			}
			else
			{
	  			fprintf(stderr, "%s: error in closing the output-file, in: %s\n", progname.c_str(), "close_in_out_files()");
			}
		}
	  ofd = 0;
	}
}
void do_exit(int exit_code)
 {
        exit(exit_code);
 }


 int do_stat(const char *name, struct stat *sbuf)
 {
   int rc;

   errno =0;
    if(!to_stdout)
    {
	rc = lstat(name, sbuf);
    }
    else
    {
	rc = stat(name, sbuf);
    }
   return rc;
 }

 int do_stat(int fd, struct stat *sbuf)
 {
   int rc;
    errno =0;
    rc = fstat(fd, sbuf);
   return rc;
 }
 	
 uInt seek_on_err(int ifd, size_t count)
 {
        uInt avail;
	off_t cur_pos; 

        if(debug == 1) 
	{
        	sprintf(log_buf, "%s: error reading input file/device, errno=%d; trying to read next block: %d bytes.", progname.c_str(), errno, io_block_size);
		log_str = log_buf;
		log(log_str);
	}
	else
	{
		fprintf(stderr, "%s: error reading input file/device, errno=%d; trying to read next block: %d bytes.\n", progname.c_str(), errno, io_block_size);
	}

	errno = 0;

	cur_pos = lseek(ifd, count, SEEK_CUR);
	avail = count;
	return avail;
 }

 int treat_file(string iname)
 {
   int rc = 0; // return value
   int zsplit_res;
   of_count = 0;	// amount of output files
   string ifn, infn; 		// name of input-file
   string ofn, outfn; 		// name of output-file
   string path_ofn; 	// constructed path-name of the output file
   string path_ifn; 	// constructed path-name of the input file
   int indx = 0; 	// index for the naming of splitted parts of output files
   int close_res; 	// result of closing zipped file
   string path_in; 	// input pathname
   string path_out; 	// output pathname

   ifn = "";
   infn = "";
   ofn = "";
   outfn = "";
   path_ofn = "";
   path_ifn = "";
   path_in = "";
   path_out = "";

   	if(!strcmp(iname.c_str(), "-")) 
	{
                from_stdin = 1;
		rc = treat_stdin();
	}
	else
	{
		if(!to_stdout)
		{ 
			if (sizeof ifname - 1 <= iname.length())
			{
				if(debug == 1) 
				{
                                        sprintf(log_buf, "%s: file name is too long: %s, in: %s", progname.c_str(), iname.c_str(), "treat_file()");
					log_str = log_buf;
					log(log_str);
				}
				else 
				{
					fprintf (stderr, "%s: file name is too long: %s, in: %s\n", progname.c_str(), iname.c_str(), "treat_file()");
				}
				rc = -1;
				return rc;
			}
				
			if(out_filename == 0 && opath == 0 && ipath == 0)
			{

				ifn = cut_path(iname);
				path_in = get_path(iname);
				
				path_ifn = path_in + ifn;
				outfn = ifn;
				ofn = make_ofname(outfn, indx);
				path_out = "";
				path_ofn = path_out + ofn;
			}
			else if(out_filename == 1 && opath == 1 && ipath == 1)
			{
				ifn.assign(cut_path(iname));
				path_in.assign(in_pathn);
				path_ifn = path_in + ifn;
				outfn = out_filen;
				ofn = make_ofname(outfn, indx);
				path_out.assign(out_pathn);
				path_out.append("/");
				path_ofn = path_out + ofn;
			}
			else if(out_filename == 1 && opath == 1 && ipath == 0)
			{
				ifn = cut_path(iname);
				path_in = get_path(iname);
				path_ifn = path_in + ifn;
				outfn = out_filen;
				ofn = make_ofname(outfn, indx);
				path_out.assign(out_pathn);
				path_out.append("/");
				path_ofn = path_out + ofn;
			}
			else if(out_filename == 1 && opath == 0 && ipath == 0)
			{
				ifn = cut_path(iname);
				path_in = get_path(iname);
				path_ifn = path_in + ifn;
				outfn = out_filen;
				ofn = make_ofname(outfn, indx);
				path_out = "";
				path_ofn = path_out + ofn;
			}
			else if(out_filename == 0 && opath == 1 && ipath == 0)
			{
				ifn = cut_path(iname);
				path_in = get_path(iname);
				path_ifn = path_in + ifn;
				outfn = ifn;
				ofn = make_ofname(outfn, indx);
				path_out.assign(out_pathn);
				path_out.append("/");
				path_ofn = path_out + ofn;
			}
			else if(out_filename == 0 && opath == 0 && ipath == 1)
			{
				ifn = cut_path(iname);
				path_in.assign(in_pathn);
				path_in.append("/");
				path_ifn = path_in + ifn;
				outfn = ifn;
				ofn = make_ofname(outfn, indx);
				path_out = "";
				path_ofn = path_out + ofn;
			}
			else if(out_filename == 0 && opath == 1 && ipath == 1)
			{
				ifn = cut_path(iname);
				path_in.assign(in_pathn);
				path_in.append("/");
				path_ifn = path_in + ifn;
				outfn = ifn;
				ofn = make_ofname(outfn, indx);
				path_out.assign(out_pathn);
				path_out.append("/");
				path_ofn = path_out + ofn;
			}
			else if(out_filename == 1 && opath == 0 && ipath == 1)
			{
				ifn = cut_path(iname);
				path_in.assign(in_pathn);
				path_in.append("/");
				path_ifn = path_in + ifn;
				outfn = out_filen;
				ofn = make_ofname(outfn, indx);
				path_out = "";
				path_ofn = path_out + ofn;
			}
			else
			{
				if(debug == 1)
				{
                                        sprintf(log_buf, "%s: input parameters were not conform, iname: %s, in_pathn: %s, out_pathn: %s, in: %s", progname.c_str(), iname.c_str(), in_pathn.c_str(), out_pathn.c_str(), "treat_file()");
					log_str = log_buf;
					log(log_str);
				}
				else
				{
       					fprintf(stderr, "%s: input parameters were not conform, iname: %s, in_pathn: %s, out_pathn: %s, in: %s\n", progname.c_str(), iname.c_str(), in_pathn.c_str(), out_pathn.c_str(), "treat_file()");
				}
				rc = -1;
				return rc;
			}

			if (do_stat(path_ifn.c_str(), &istat) != 0)
			{
				if(debug == 1)
				{
                                        sprintf(log_buf, "%s: cannot access file/device: %s, in: %s", progname.c_str(), iname.c_str(), "treat_file()");
					log_str = log_buf;
					log(log_str);
				}
				else 
                                {
					fprintf (stderr, "%s: cannot access file/device: %s, in: %s\n", progname.c_str(), iname.c_str(), "treat_file()");
				}
				rc = -1;
				return rc;
			}


			ifd = open(path_ifn.c_str(), O_RDONLY | O_LARGEFILE);
			
			if( ifd == -1)
			{
				if(debug == 1)
				{
                                        sprintf(log_buf, "%s: cannot open file/device: %s, in: %s", progname.c_str(), path_ifn.c_str(), "treat_file()");
					log_str = log_buf;
					log(log_str);
				}
				else
                                {
					fprintf (stderr, "%s: cannot open file/device: %s, in: %s\n", progname.c_str(), path_ifn.c_str(), "treat_file()");
				}
				rc = -1;
				return rc;
			}
			
			ofd = open(path_ofn.c_str(),O_RDWR | O_TRUNC | O_CREAT | O_LARGEFILE, S_IRUSR | S_IWUSR /*| S_IRWXU */);
			
			if( ofd == -1)
			{
				if(debug == 1) 
				{
                                        sprintf(log_buf, "%s: cannot open file/device: %s, in: %s", progname.c_str(), path_ofn.c_str(), "treat_file()");
					log_str = log_buf;
					log(log_str);
				}
				else
                                {
					fprintf (stderr, "%s: cannot open file/device: %s, in: %s\n", progname.c_str(), path_ofn.c_str(), "treat_file()");
				}
				rc = -1;
				return rc;
			}

			for ( ; ; )
			{	
				zsplit_res = do_zsplit(ifd, ofd, level, path_ifn, path_ofn, chunk_size);
	
				if(zsplit_res == 1)
				{
					close_res = close(ofd);
					ofd = 0;
					if(close_res == 0)
					{
						indx++;
						ofn = make_ofname(outfn, indx);
						path_ofn = path_out + ofn;
						ofd = open(path_ofn.c_str(),O_RDWR | O_TRUNC | O_CREAT | O_LARGEFILE, S_IRUSR | S_IWUSR);
						
						if( ofd == -1)
						{
							if(debug == 1)
							{
								sprintf(log_buf, "%s: cannot open file/device: %s, in: %s", progname.c_str(), path_ofn.c_str(), "treat_file()");
								log_str = log_buf;
								log(log_str);
							}
							else
							{
								fprintf (stderr, "%s: cannot open file/device: %s, in: %s\n", progname.c_str(), path_ofn.c_str(), "treat_file()");
							}
							rc = -1;
							return rc;
						}
					}
					else
					{
						if(debug == 1)
						{
                                                        sprintf(log_buf, "%s: error in closing the file: %s, in: %s", progname.c_str(), path_ofn.c_str(), "treat_file()");
							log_str = log_buf;
							log(log_str);
						}
						else
						{
							fprintf (stderr, "%s: error in closing the file: %s, in: %s\n", progname.c_str(), path_ofn.c_str(), "treat_file()");
						}
						close(ifd);
						ifd = 0;
						rc = -1;
						break;
					}
				}
	
				else if(zsplit_res == 0)
				{
					close_in_out_files(ifd, ofd);
					rc = 0;
					break;
				}
				else if(zsplit_res == -2)
				{
                                        if(debug == 1)
					{
                                                sprintf(log_buf, "%s: error in deflate function, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
						log_str = log_buf;
						log(log_str);
					}
					else
					{
						fprintf (stderr, "%s: error in deflate function, in: %s return=%d\n", progname.c_str(), "do_zsplit()", zsplit_res);
					}
					close_in_out_files(ifd, ofd);
					rc = -1;
					break;
				}
				else if(zsplit_res == -3)
				{
                                        if(debug == 1)
					{
                                                sprintf(log_buf, "%s: error in estimating the file size in do_stat(), in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
						log_str = log_buf;
						log(log_str);
					}
					else
					{
						fprintf (stderr, "%s: error in estimating the file size in do_stat(), in: %s return=%d\n", progname.c_str(), "do_zsplit()", zsplit_res);
					}
					close_in_out_files(ifd, ofd);
					rc = -1;
					break;
				}
				else if(zsplit_res == -5)
				{
					if(debug == 1)
					{
                                                sprintf(log_buf, "%s: error in writing output file, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
						log_str = log_buf;
						log(log_str);
					}
					else
					{
						fprintf (stderr, "%s: error in writing output file, in: %s return=%d\n", progname.c_str(), "do_zsplit()", zsplit_res);
					}
					close_in_out_files(ifd, ofd);
					rc = -1;
					break;
				}
				else if(zsplit_res == -6)
				{
					if(debug == 1)
					{
                                                sprintf(log_buf, "%s: error in the initialization of deflation, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
						log_str = log_buf;
						log(log_str);
					}
					else
                                        {
						fprintf (stderr, "%s: error in the initialization of deflation, in: %s return=%d\n", progname.c_str(), "do_zsplit()", zsplit_res);
					}
					close_in_out_files(ifd, ofd);
					rc = -1;
					break;
				}
				else if(zsplit_res == -7)
				{
                                        if(debug == 1)
					{
                                                sprintf(log_buf, "%s: unknown error, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
						log_str = log_buf;
						log(log_str);
					}
					else
                                        {
						fprintf (stderr, "%s: unknown error, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
					}
					close_in_out_files(ifd, ofd);
					rc = -1;
					break;
				}
				else if(zsplit_res == -8)
				{
                                        if(debug == 1)
					{
                                                sprintf(log_buf, "%s: deflate(): stream state was inconsistent, next_in or next_out was NULL, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
						log_str = log_buf;
						log(log_str);
					}
					else
					{
						fprintf (stderr, "%s: deflate(): stream state was inconsistent, next_in or next_out was NULL, in: %s return=%d\n", progname.c_str(), "do_zsplit()", zsplit_res);
					}
					close_in_out_files(ifd, ofd);
					rc = -1;
					break;
				}
				else if(zsplit_res == -9)
				{
                                        if(debug == 1)
					{
                                                sprintf(log_buf, "%s: deflate(): no progress was possible, avail_in or avail_out was NULL, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
						log_str = log_buf;
						log(log_str);
					}
					else 
					{
						fprintf (stderr, "%s: deflate(): no progress was possible, avail_in or avail_out was NULL, in: %s return=%d\n", progname.c_str(), "do_zsplit()", zsplit_res);
					}
					close_in_out_files(ifd, ofd);
					rc = -1;
					break;
				}
				else
				{
                                        if(debug == 1) 
					{
                                                sprintf(log_buf, "%s: unexpected error..., in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
						log_str = log_buf;
						log(log_str);
					}
					else 
					{
						fprintf (stderr, "%s: unexpected error..., in: %s return=%d\n", progname.c_str(), "do_zsplit()", zsplit_res);
					}
					close_in_out_files(ifd, ofd);
					rc = -1;
					break;
				}
			}
		} 
		else
		{ 

			ofn.assign("stdout");
			if (do_stat(fileno(stdout), &istat) != 0)
			{
                                if(debug == 1) 
				{
                                        sprintf(log_buf, "%s: cannot acess file/device: %s, in: %s", progname.c_str(), ofn.c_str(), "treat_file()");
					log_str = log_buf;
					log(log_str);
				}
				else 
              			{
					fprintf (stderr, "%s: cannot acess file/device: %s, in: %s\n", progname.c_str(), ofn.c_str(), "treat_file()");
				}
				rc = -1;
				return rc;
			}

			
			ofd = fileno(stdout);

			if (sizeof ifname - 1 <= iname.length())
			{
                                if(debug == 1)
				{
                                        sprintf(log_buf, "%s: file name too long: %s, in: %s", progname.c_str(), iname.c_str(), "treat_file()");
					log_str = log_buf;
					log(log_str);
				}
				else
				{
					fprintf (stderr, "%s: file name too long: %s, in: %s\n", progname.c_str(), iname.c_str(), "treat_file()");
				}
				rc = -1;
				return rc;
			}
			
			if (do_stat(iname.c_str(), &istat) != 0)
			{
                                if(debug == 1)
				{
                                        sprintf(log_buf, "%s: cannot acess file/device: %s, in: %s", progname.c_str(), iname.c_str(), "treat_file()");
					log_str = log_buf;
					log(log_str);
				}
				else 
				{
					fprintf (stderr, "%s: cannot acess file/device: %s, in: %s", progname.c_str(), iname.c_str(), "treat_file()");
				}
				rc = -1;
				return rc;
			}

			if(ipath == 0)
			{
				ifn = cut_path(iname);
				path_in = get_path(iname);
				path_ifn = path_in + ifn;
			}
			if(ipath == 1)
			{
				ifn = cut_path(iname);
				path_in.assign(in_pathn);
				path_in.append("/");
				path_ifn = path_in + ifn;
			}

			ifd = open(path_ifn.c_str(), O_RDONLY | O_LARGEFILE, S_IRUSR | S_IWUSR);
			
			if( ifd == -1)
			{
				if(debug == 1)
				{
                                        sprintf(log_buf, "%s: cannot open file/device: %s, in: %s", path_ifn.c_str(), path_ifn.c_str(), "treat_file()");
					log_str = log_buf;
					log(log_str);
				}
				else
                                {
					fprintf (stderr, "%s: cannot open file/device: %s, in: %s\n", path_ifn.c_str(), path_ifn.c_str(), "treat_file()");
				}
				rc = -1;
				return rc;
			}

			for ( ; ; )
			{
				zsplit_res = do_zsplit(ifd, ofd, level, path_ifn, path_ofn);

				if(zsplit_res == 0)
				{
					close_in_out_files(ifd, ofd);
					rc = 0;
					break;
				}
				else if(zsplit_res == -2)
				{
                                        if(debug == 1)
					{
                                                sprintf(log_buf, "%s: error in deflate function for stdout, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
						log_str = log_buf;
						log(log_str);
					}
					else
					{
						fprintf (stderr, "%s: error in deflate function for stdout, in: %s return=%d\n", progname.c_str(), "do_zsplit()", zsplit_res);
					}
					close_in_out_files(ifd, ofd);
					rc = -1;
					break;
				}
				else if(zsplit_res == -4)
				{
                                        if(debug == 1)
					{
                                                sprintf(log_buf, "%s: error in accessing stdin in do_stat() for stdout, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
						log_str = log_buf;
						log(log_str);
					}
					else
					{
						fprintf (stderr, "%s: error in accessing stdin in do_stat() for stdout, in: %s return=%d\n", progname.c_str(), "do_zsplit()", zsplit_res);
					}
					close_in_out_files(ifd, ofd);
					rc = -1;
					break;
				}
				else if(zsplit_res == -5)
				{
                                        if(debug == 1)
					{
                                                sprintf(log_buf, "%s: error in writing output to stdout, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
						log_str = log_buf;
						log(log_str);
					}
					else
                                        {
						fprintf (stderr, "%s: error in writing output to stdout, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
					}
					close_in_out_files(ifd, ofd);
					rc = -1;
					break;
				}
				else if(zsplit_res == -6)
				{
                                        if(debug == 1)
					{
                                                sprintf(log_buf, "%s: error in initialization of deflation for stdout, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
						log_str = log_buf;
						log(log_str);
					}
					else
					{
						fprintf (stderr, "%s: error in initialization of deflation for stdout, in: %s return=%d\n", progname.c_str(), "do_zsplit()", zsplit_res);
					}
					close_in_out_files(ifd, ofd);
					rc = -1;
					break;
				}
				else if(zsplit_res == -7)
				{
                                        if(debug == 1)
					{
                                                sprintf(log_buf, "%s: unknown error... for stdout, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
						log_str = log_buf;
						log(log_str);
					}
					else 
					{
						fprintf (stderr, "%s: unknown error... for stdout, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
					}
					close_in_out_files(ifd, ofd);
					rc = -1;
					break;
				}
				else if(zsplit_res == -8)
				{
                                        if(debug == 1)
					{
                                                sprintf(log_buf, "%s: deflate(): stream state was inconsistent next_in or next_out was NULL for stdout, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
						log_str = log_buf;
						log(log_str);
					}
					else
					{
						fprintf (stderr, "%s: deflate(): stream state was inconsistent next_in or next_out was NULL for stdout, in: %s return=%d\n", progname.c_str(), "do_zsplit()", zsplit_res);
					}
					close_in_out_files(ifd, ofd);
					rc = -1;
					break;
				}
				else if(zsplit_res == -9)
				{
                                        if(debug == 1)
					{
                                                sprintf(log_buf, "%s: deflate(): no progress was possible, avail_in or avail_out was NULL for stdout, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
						log_str = log_buf;
						log(log_str);
					}
					else
					{
						fprintf (stderr, "%s: deflate(): no progress was possible, avail_in or avail_out was NULL for stdout, in: %s return=%d\n", progname.c_str(), "do_zsplit()", zsplit_res);
					}
					close_in_out_files(ifd, ofd);
					rc = -1;
					break;
				}
				else if(zsplit_res == -10)
				{
                                        if(debug == 1)
					{
                                                sprintf(log_buf, "%s: error in accessing stdout in do_stat(), in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
						log_str = log_buf;
						log(log_str);
					}
					else
					{
						fprintf (stderr, "%s: error in accessing stdout in do_stat(), in: %s return=%d\n", progname.c_str(), "do_zsplit()", zsplit_res);
					}
					close_in_out_files(ifd, ofd);
					rc = -1;
					break;
				}
				else
				{
                                        if(debug == 1)
					{
                                                sprintf(log_buf, "%s: unexpected error..., for stdout, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
						log_str = log_buf;
						log(log_str);
					}
					else
					{
						fprintf (stderr, "%s: unexpected error..., for stdout, in: %s return=%d\n", progname.c_str(), "do_zsplit()", zsplit_res);
					}
					close_in_out_files(ifd, ofd);
					rc = -1;
					break;
				}
			}

		}

	}
  total_time_elps = elaps_time(end_sum_elps_time);
  
  av_speed_s = av_speed(end_sum_elps_time, end_sum_read);
  
  if(debug == 1)
  {
	sprintf(log_buf, "%s: total sum of bytes read: %lld; written: %lld; elapsed time: %s", progname.c_str(), end_sum_read, end_sum_written, total_time_elps.c_str());
	log_str = log_buf;
	log(log_str);
	
	sprintf(log_buf, "%s: average speed of imaging: %s", progname.c_str(), av_speed_s.c_str());
	log_str = log_buf;
	log(log_str);
  }
  else
  {
  	fprintf (stderr, "%s: total sum of bytes read: %lld; written: %lld; elapsed time: %s\n", progname.c_str(), end_sum_read, end_sum_written, total_time_elps.c_str());
	fprintf (stderr, "%s: average speed of imaging: %s\n", progname.c_str(), av_speed_s.c_str());
  }
  return rc;
 }

 int do_zsplit(int ifd, int ofd, int level, string path_ifn, string path_ofn, off_t chnk_size)
{
 int res = 0;
 int count = 0; 
 z_stream z;
 Bytef input_buffer[io_block_size];
 Bytef output_buffer[io_block_size];
 int status = Z_OK;
 size_t read_res = 0;	// result of read operation could be SAFE_READ_ERROR (-1)
 struct stat ostat;         	/* status for output file */
 int testwr =0;
 unsigned long long sum_read = 0;	// the ammount of read bytes
 unsigned long long sum_written = 0; // the amount of written bytes
 
 unsigned long sum_elps_time = 0;  	// elapsed time in seconds 
 time_t start_sec, end_sec; 	// begin time im seconds, end time in seconds
 string time_elps; 		// elapsed time as formatted string	
 			





        z.zalloc = (alloc_func)0;
	z.zfree = (free_func)0;
	z.opaque = (voidpf)0;

	z.avail_in = 0;
	z.next_out = output_buffer;
	z.avail_out = io_block_size;

	if(deflateInit(&z, level) != Z_OK)
	{
		res = -6;
		return res;
	}
	
	time(&start_sec);

	for ( ; ; )
	{
		if (z.avail_in == 0)
		{

			if(noerror == 1)
			{
				memset(input_buffer, '\0', io_block_size);
			}

		  	z.next_in = input_buffer;
		 	read_res = safe_read(ifd, input_buffer, io_block_size);
			if(read_res == SAFE_READ_ERROR)
			{
				z.avail_in = 0;
			}
			else
			{
                               z.avail_in = read_res;
			}
                 
			if(read_res == SAFE_READ_ERROR && noerror == 1)
			{
				z.avail_in = seek_on_err(ifd, io_block_size);
			} 

         	  sum_read += z.avail_in;
		}

		if (z.avail_in == 0)
		{
		  z.next_out = output_buffer;
		  z.avail_out = io_block_size;
		  res = 0;
		  break;
		}

		status = deflate(&z, Z_NO_FLUSH);
		if (status != Z_OK)
		{
		  res = -2;
		  break;
		}

		if(status == Z_OK && z.avail_out == 0)
		{
			do
			{
			  count = io_block_size - z.avail_out;
			  if(count)
			  {
			  	testwr = safe_write(ofd, output_buffer, count);
				sum_written += testwr; 
		  	  	if(testwr != count)
		  	  	{
    		  			res = -5; 
    		  			break;
		  	  	}
				z.next_out = output_buffer;
				z.avail_out = io_block_size;
			   }
			   status = deflate(&z, Z_NO_FLUSH);
			} while (z.avail_out == 0);

			count = io_block_size - z.avail_out;
			if(count)
			{
				testwr = safe_write(ofd, output_buffer, count);
				sum_written += testwr;
		  	  	if(testwr != count)
		  	  	{
    		  			res = -5;
    		  			break;
		  	  	}
			}
		}
		else if(status == Z_OK && z.avail_out == io_block_size)
		{
			do
			{
                                if(noerror == 1)
				{
					memset(input_buffer, '\0', io_block_size);
				}
				z.next_in = input_buffer;
		  	  	read_res = safe_read(ifd, input_buffer, io_block_size);
				if(read_res == SAFE_READ_ERROR)
				{
					z.avail_in = 0;
				}
				else
				{
                               		z.avail_in = read_res;
				}
				if(read_res == SAFE_READ_ERROR && noerror == 1)
				{
					z.avail_in = seek_on_err(ifd, io_block_size);
				} 
				sum_read += z.avail_in;
			  status = deflate(&z, Z_SYNC_FLUSH);
			} while (z.avail_out == io_block_size);

			count = io_block_size - z.avail_out;

			if(count)
			{
		  		testwr = safe_write(ofd, output_buffer, count);
				sum_written += testwr;
		  		if(testwr != count)
		  		{
    		  			res = -5;
    		  			break;
		  		}
			}
		}
		else if(status == Z_OK && (z.avail_out < 0 || z.avail_out < io_block_size))
		{
			count = io_block_size - z.avail_out;

			if(count)
			{
		  		testwr = safe_write(ofd, output_buffer, count);
				sum_written += testwr;
		  		if(testwr != count)
		  		{
    		  			res = -5;
    		  			break;
		  		}
			}
		}
		else
		{
		  res = -7;
		  break;
		}

    		if (do_stat(path_ofn.data(), &ostat) != 0)
		{
    		  res = -3;
    		  break;
		}

		if (chunk_size_known == 1 && ostat.st_size >= chnk_size)
		{
		  z.next_out = output_buffer;
		  z.avail_out = io_block_size;
		  res = 1;
		  break;
		}
		z.next_out = output_buffer;
		z.avail_out = io_block_size;
	}
	for ( ; ; )
	{
		status = deflate(&z, Z_FINISH);
		if(status == Z_OK)
		{
			do
			{
				count = io_block_size -z.avail_out;
				if(count)
				{
					testwr = safe_write(ofd, output_buffer, count);
					sum_written += testwr;
					if(testwr != count)
		  			{
    		  				res = -5;
    		  				break;
		  			}
					z.next_out = output_buffer;
					z.avail_out = io_block_size;
				}

				status = deflate(&z, Z_FINISH);

			}while (status == Z_OK);

			count = io_block_size - z.avail_out;

			if(count)
			{
				testwr = safe_write(ofd, output_buffer, count);
				sum_written += testwr;
				if(testwr != count)
		  		{
    		  			res = -5;
    		  			break;
		  		}

			}
		  break;
		}
		else if(status == Z_STREAM_END)
		{
                	count = io_block_size - z.avail_out;

			if(count)
			{
				testwr = safe_write(ofd, output_buffer, count);
				sum_written += testwr;
				if(testwr != count)
		  		{
    		  			res = -5;
    		  			break;
		  		}

			}
		  break;
		}
               	else if (status == Z_STREAM_ERROR)
		{
			res = -8; 
			break;
		}
		else if (status == Z_BUF_ERROR)
		{
			break;
		}
	}   	
  deflateEnd(&z);
  
  time(&end_sec);
    
 sum_elps_time = (unsigned long)end_sec - (unsigned long)start_sec;
 time_elps = elaps_time(sum_elps_time);
  
  
  if(debug == 1)
  {
        sprintf(log_buf, "%s: bytes read from %s: %lld; written to %s: %lld; elapsed time: %s", progname.c_str(), path_ifn.c_str(), sum_read, path_ofn.c_str(), sum_written, time_elps.c_str());
	log_str = log_buf;
	log(log_str);
  }
  else
  {
  	fprintf (stderr, "%s: bytes read from %s: %lld; written to %s: %lld; elapsed time: %s\n", progname.c_str(), path_ifn.c_str(), sum_read, path_ofn.c_str(), sum_written, time_elps.c_str());
  }
  
  end_sum_read += sum_read;
  end_sum_written += sum_written;
  end_sum_elps_time += sum_elps_time;

 return res;
}

int do_zsplit(int ifd, int ofd, int level, string path_ifn, string path_ofn)
{
 int res = 0;
 int count;  // the size of the buffer after compression
 z_stream z;
 Bytef input_buffer[io_block_size];
 Bytef output_buffer[io_block_size];
 int status = Z_OK;
 struct stat ostat;         	/* status for output file */
 int testwr =0;
 unsigned long long sum_read = 0;	// the ammount of read bytes
 unsigned long long sum_written = 0; // the amount of written bytes
 string logstr;
 size_t read_res = 0;	// result of read operation could be SAFE_READ_ERROR (-1)

 unsigned long sum_elps_time = 0;  	// elapsed time in seconds 
 time_t start_sec, end_sec; 	// begin time im seconds, end time in seconds
 string time_elps; 		// elapsed time as formatted string	


 	z.zalloc = (alloc_func)0;
	z.zfree = (free_func)0;
	z.opaque = (voidpf)0;

	z.avail_in = 0;
	z.next_out = output_buffer;
	z.avail_out = io_block_size;

	if(deflateInit(&z, level) != Z_OK)
	{
		res = -6;
		return res;
	}

	time(&start_sec);

	for ( ; ; )
	{
		if (z.avail_in == 0)
		{
		  if(noerror == 1)
		  {
			memset(input_buffer, '\0', io_block_size);
		  }
		  z.next_in = input_buffer;
		  read_res = safe_read(ifd, input_buffer, io_block_size);
		  if(read_res == SAFE_READ_ERROR)
		  {
			z.avail_in = 0;
		  }
		  else
		  {
                       z.avail_in = read_res;
		  }
		  
		  if(read_res == SAFE_READ_ERROR && noerror == 1)
		  {
			z.avail_in = seek_on_err(ifd, io_block_size);
		  }
		  sum_read += z.avail_in;
		}

		if (z.avail_in == 0)
		{
		  z.next_out = output_buffer;
		  z.avail_out = io_block_size;
		  res = 0;
		  break;
		}

		status = deflate(&z, Z_NO_FLUSH);
		if (status != Z_OK)
		{
		  res = -2;
		  break;
		}

		if(status == Z_OK && z.avail_out == 0)
		{
			do
			{
			  count = io_block_size - z.avail_out;
			  if(count)
			  {
			  	testwr = safe_write(ofd, output_buffer, count);
				sum_written += testwr;
		  	  	if(testwr != count)
		  	  	{
    		  			res = -5;
    		  			break;
		  	  	}
				z.next_out = output_buffer;
				z.avail_out = io_block_size;
			   }
			   status = deflate(&z, Z_NO_FLUSH);
			} while (z.avail_out == 0);

			count = io_block_size - z.avail_out;
			if(count)
			{
				testwr = safe_write(ofd, output_buffer, count);
				sum_written += testwr;
		  	  	if(testwr != count)
		  	  	{
    		  			res = -5;
    		  			break;
		  	  	}
			}
		}
		else if(status == Z_OK && z.avail_out == io_block_size)
		{
			do
			{
                                if(noerror == 1)
				{
					memset(input_buffer, '\0', io_block_size);
				}
				z.next_in = input_buffer;
		  	  	read_res = safe_read(ifd, input_buffer, io_block_size);
				if(read_res == SAFE_READ_ERROR)
				{
					z.avail_in = 0;
				}
				else
				{
                               		z.avail_in = read_res;
				}
             
				if(read_res == SAFE_READ_ERROR && noerror == 1)
				{
					z.avail_in = seek_on_err(ifd, io_block_size);
				} 
				sum_read += z.avail_in;
			  status = deflate(&z, Z_SYNC_FLUSH);
			} while (z.avail_out == io_block_size);

			count = io_block_size - z.avail_out;

			if(count)
			{
		  		testwr = safe_write(ofd, output_buffer, count);
				sum_written += testwr;
		  		if(testwr != count)
		  		{
    		  			res = -5;
    		  			break;
		  		}
			}
		}
		else if(status == Z_OK && (z.avail_out < 0 || z.avail_out < io_block_size))
		{
			count = io_block_size - z.avail_out;

			if(count)
			{
		  		testwr = safe_write(ofd, output_buffer, count);
				sum_written += testwr;
		  		if(testwr != count)
		  		{
    		  			res = -5;
    		  			break;
		  		}
			}
		}
		else
		{
		  res = -7;
		  break;
		}

    		if (do_stat(fileno(stdout), &ostat) != 0)
		{
    		  res = -10;
    		  break;
		}
		
		if(do_stat(fileno(stdin), &istat) != 0)
		{
    		  res = -4;
    		  break;
		}

		z.next_out = output_buffer;
		z.avail_out = io_block_size;
	}
	for ( ; ; )
	{
		status = deflate(&z, Z_FINISH);
		if(status == Z_OK)
		{
			do
			{
				count = io_block_size -z.avail_out;
				if(count)
				{
					testwr = safe_write(ofd, output_buffer, count);
					sum_written += testwr;
					if(testwr != count)
		  			{
    		  				res = -5;
    		  				break;
		  			}
					z.next_out = output_buffer;
					z.avail_out = io_block_size;
				}

				status = deflate(&z, Z_FINISH);

			}while (status == Z_OK);

			count = io_block_size - z.avail_out;

			if(count)
			{
				testwr = safe_write(ofd, output_buffer, count);
				sum_written += testwr;
				if(testwr != count)
		  		{
    		  			res = -5;
    		  			break;
		  		}
			}
		  break;
		}
		else if(status == Z_STREAM_END)
		{
                	count = io_block_size - z.avail_out;

			if(count)
			{
				testwr = safe_write(ofd, output_buffer, count);
				sum_written += testwr;
				if(testwr != count)
		  		{
    		  			res = -5;
    		  			break;
		  		}

			}
		  break;
		}
               	else if (status == Z_STREAM_ERROR)
		{
			res = -8; 
			break;
		}
		else if (status == Z_BUF_ERROR)
		{
			break;
		}
	}

  deflateEnd(&z);
  
  time(&end_sec);
  
 sum_elps_time = (unsigned long)end_sec - (unsigned long)start_sec;
 time_elps = elaps_time(sum_elps_time);
 

  if(debug == 1)
  {
        sprintf(log_buf, "%s: bytes read from %s: %lld; written to %s: %lld; elapsed time: %s", progname.c_str(), path_ifn.c_str(), sum_read, path_ofn.c_str(), sum_written, time_elps.c_str());
	log_str = log_buf;
	log(log_str);
  }
  else
  {
  	fprintf (stderr, "%s: bytes read from %s: %lld; written to %s: %lld; elapsed time: %s\n", progname.c_str(), path_ifn.c_str(), sum_read, path_ofn.c_str(), sum_written, time_elps.c_str());
  }

  end_sum_read += sum_read;
  end_sum_written += sum_written;
  end_sum_elps_time += sum_elps_time;

 return res;
}

int treat_stdin()
{
  int res = 0;
  int rc = 0; // return value for the main function
  string path_ifn, infn; 		// names of input-file
  string path_ofn, ofn, outfn; 		// names of output-file
  int indx = 0;			// index for the creation of output files
  string path_out;
  int zsplit_res;	// return result of the zsplit function
  int close_res; 	// result of closing zipped file

  path_ifn.assign("stdin");

  if (do_stat(fileno(stdin), &istat) != 0)
  {
        if(debug == 1)
	{
                sprintf(log_buf, "%s: cannot access file/device: %s, in: %s", progname.c_str(), path_ifn.c_str(), "treat_stdin()");
		log_str = log_buf;
		log(log_str);
	}
	else 
	{
		fprintf (stderr, "%s: cannot access file/device: %s, in: %s\n", progname.c_str(), path_ifn.c_str(), "treat_stdin()");
	}
	res = -1;
	return res;
  }
  
  ifd = fileno(stdin);

  if(to_stdout)
  {
  	path_ofn.assign("stdout");
        if (do_stat(fileno(stdout), &istat) != 0)
	{
		if(debug == 1)
		{
                        sprintf(log_buf, "%s: cannot access file/device: %s, in: %s", progname.c_str(), path_ofn.c_str(), "treat_stdin()");
			log_str = log_buf;
			log(log_str);
		}
		else
		{
			fprintf (stderr, "%s: cannot access file/device: %s, in: %s\n", progname.c_str(), path_ofn.c_str(), "treat_stdin()");
		}
		res = -1;
		return res;
	}

	ofd = fileno(stdout);

	for ( ; ; )
  	{
		zsplit_res = do_zsplit(ifd, ofd, level, path_ifn, path_ofn);

		if(zsplit_res == 0)
		{
			close_in_out_files(ifd, ofd);
			res = 0;
			break;
		}
		else if(zsplit_res == -2)
		{
                        if(debug == 1)
			{
                                sprintf(log_buf, "%s: error in deflate function for stdout, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
				log_str = log_buf;
				log(log_str);
			}
			else
			{
				fprintf (stderr, "%s: error in deflate function for stdout: %s: %s: %d\n", progname.c_str(), "do_zsplit()", zsplit_res);
			}
			close_in_out_files(ifd, ofd);
			res = -1;
			break;
		}
		else if(zsplit_res == -4)
		{
                        if(debug == 1)
			{
                                sprintf(log_buf, "%s: error in accessing stdin in: do_stat(), for stdout, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
				log_str = log_buf;
				log(log_str);
			}
			else
			{
				fprintf (stderr, "%s: error in accessing stdin in: do_stat(), for stdout, in: %s return=%d\n", progname.c_str(), "do_zsplit()", zsplit_res);
			}
			close_in_out_files(ifd, ofd);
			res = -1;
			break;
		}
		else if(zsplit_res == -10)
		{
                        if(debug == 1)
			{
                                sprintf(log_buf, "%s: error in accessing stdout in: do_stat(), for stdout, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
				log_str = log_buf;
				log(log_str);
			}
			else
			{
				fprintf (stderr, "%s: error in accessing stdout in: do_stat(), for stdout, in: %s return=%d\n", progname.c_str(), "do_zsplit()", zsplit_res);
			}
			close_in_out_files(ifd, ofd);
			res = -1;
			break;
		}
		else
		{
                        if(debug == 1)
			{
                                sprintf(log_buf, "%s: unexpected error..., for stdout, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
				log_str = log_buf;
				log(log_str);
			}
			else
			{
				fprintf (stderr, "%s: unexpected error..., for stdout, in: %s return=%d\n", progname.c_str(), "do_zsplit()", zsplit_res);
			}
			close_in_out_files(ifd, ofd);
			res = -1;
			break;
		}
	}
  }
  else 
  {
  	if(out_filename == 0 && opath == 0)
	{
		outfn = get_def_filename();
		ofn = make_ofname(outfn, indx);
		path_out = "";
		path_ofn = path_out + ofn;
	}
	else if(out_filename == 1 && opath == 1)
	{
		outfn = out_filen;
		ofn = make_ofname(outfn, indx);
		path_out.assign(out_pathn);
		path_out.append("/");
		path_ofn = path_out + ofn;
	}
	else if(out_filename == 1 && opath == 0)
	{
		outfn = out_filen;
		ofn = make_ofname(outfn, indx);
		path_out = "";
		path_ofn = path_out + ofn;
	}
	else if(out_filename == 0 && opath == 1)
	{
		outfn = get_def_filename();
		ofn = make_ofname(outfn, indx);
		path_out.assign(out_pathn);
		path_out.append("/");
		path_ofn = path_out + ofn;
	}
	else
	{
                if(debug == 1)
		{
                        sprintf(log_buf, "%s: input parameters were not conform, path_ifn: %s, path_ofn: %s, in: %s", progname.c_str(), path_ifn.c_str(), path_ofn.c_str(), "do_zsplit()");
			log_str = log_buf;
			log(log_str);
		}
		else
		{
			fprintf(stderr, "%s: input parameters were not conform, path_ifn: %s, path_ofn: %s, in: %s\n", progname.c_str(), path_ifn.c_str(), path_ofn.c_str(), "do_zsplit()");
		}
		res = -1;
		return res;
	}

	ofd = open(path_ofn.c_str(),O_RDWR | O_TRUNC | O_CREAT | O_LARGEFILE, S_IRUSR | S_IWUSR);
	
	if( ofd == -1)
	{
		if(debug == 1)
		{
			sprintf(log_buf, "%s: cannot open file/device: %s, in: %s", path_ifn.c_str(), path_ofn.c_str(), "treat_file()");
			log_str = log_buf;
			log(log_str);
		}
		else
		{
			fprintf (stderr, "%s: cannot open file/device: %s, in: %s\n", path_ifn.c_str(), path_ofn.c_str(), "treat_file()");
		}
		rc = -1;
		return rc;
	}

	for ( ; ; )
  	{
		zsplit_res = do_zsplit(ifd, ofd, level, path_ifn, path_ofn, chunk_size);

		if(zsplit_res == 1)
		{
			close_res = close(ofd);
			ofd = 0;
			if(close_res == 0)
			{
				indx++;
				ofn = make_ofname(outfn, indx);
				path_ofn = path_out + ofn;
				ofd = open(path_ofn.c_str(),O_RDWR | O_TRUNC | O_CREAT | O_LARGEFILE, S_IRUSR | S_IWUSR);
				if( ofd == -1)
				{
					if(debug == 1)
					{
						sprintf(log_buf, "%s: cannot open file/device: %s, in: %s", path_ofn.c_str(), path_ofn.c_str(), "treat_file()");
						log_str = log_buf;
						log(log_str);
					}
					else
					{
						fprintf (stderr, "%s: cannot open file/device: %s, in: %s\n", path_ofn.c_str(), path_ofn.c_str(), "treat_file()");
					}
					rc = -1;
					return rc;
				}
			}
			else
			{
                                if(debug == 1)
				{
                                        sprintf(log_buf, "%s: error in closing the file, in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
					log_str = log_buf;
					log(log_str);
				}
				else
				{
					fprintf (stderr, "%s: error in closing the file, in: %s return=%d\n", progname.c_str(), "do_zsplit()", zsplit_res);
				}
				close_in_out_files(ifd, ofd);
				res = -1;
				break;
			}
		}
		else if(zsplit_res == 0)
		{
			close_in_out_files(ifd, ofd);
			res = 0;
			break;
		}
		else if(zsplit_res == -2)
		{
                        if(debug == 1)
			{
                                sprintf(log_buf, "%s: error in accessing stdout in: do_stat(), in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
				log_str = log_buf;
				log(log_str);
			}
			else
			{
				fprintf (stderr, "%s: error in accessing stdout in: do_stat(), in: %s return=%d\n", progname.c_str(), "do_zsplit()", zsplit_res);
			}
			close_in_out_files(ifd, ofd);
			res = -1;
			break;
		}
		else if(zsplit_res == -3)
		{
                        if(debug == 1)
			{
                                sprintf(log_buf, "%s: error in estimating the file size in: do_stat(), in: %s return=%d", progname.c_str(), "do_zsplit()", zsplit_res);
				log_str = log_buf;
				log(log_str);
			}
			else
			{
				fprintf (stderr, "%s: error in estimating the file size in: do_stat(), in: %s return=%d\n", progname.c_str(), "do_zsplit()", zsplit_res);
			}
			close_in_out_files(ifd, ofd);
			res = -1;
			break;
		}
  	}
  }
  return res;
}
